#include "comfio.h"
#include "ipc_exception.h"
#include <iostream>
#include <cstdlib>

extern "C" {
	#include <sys/time.h>
	#include <errno.h>
}


/*! 
 					constructor for ComfIO
					in: 
						- bufsize: size of input buffer 
*/

ComfIO::ComfIO(ssize_t bufsize) : eots_call_ptr(NULL), bytes_to_send(0), atomic(false), cons_debug(false)
{
	comfiobuf 		= new char[max_msg = bufsize];
	eots			= new string("");
	ioctrls_chars	= NULL;
}

/*! 
 					destructor for ComfIO, frees buffer memory

					Can throw:
						- IpcException
*/


ComfIO::~ComfIO()
{
	if(atomic)
	{
		lockMutex("~ComfIO");
		unlockMutex("~ComfIO");
		if(pthread_mutex_destroy(&amutex) != 0)
		{
			cerr << "Mutex destroy error: " << ::strerror(errno);
		}
	}
	delete[] comfiobuf;
	delete eots;
	if(cons_debug)
	{
		delete[] ioctrls_chars;
	}
}


/*!
		initializes mutex used with lockMutex / unlockMutex

		in:
			- true : the mutex will be initialized
			
			-false : the mutex will be destroyed

		Can throw:
			- IpcException

			(the mutex will be destroyed at deletion, if atomic is true)
				
*/

void ComfIO::atomicIO(bool on) 
{
	atomic = on;
	stringstream s;
	if(on)
	{
		if(pthread_mutex_init(&amutex, 0) != 0)
		{
			s << "Mutex init error: " << ::strerror(errno);
			throw IpcException(s.str(), __FILE__, __LINE__);
		}
	}
	else
	{
		if(pthread_mutex_destroy(&amutex) != 0)
		{
			s << "Mutex destroy error: " << ::strerror(errno);
			throw IpcException(s.str(), __FILE__, __LINE__);
		}
	}
}

ssize_t ComfIO::count_chars(const char*s, char c)
{
	ssize_t count = 0;
	for(;*s;s++)
	{
		if(*s == c)
		{
			++count;
		}
	}
	return count;
}

/*!
	lock mutex if atomicIO() is true

	Can throw:
						- IpcException
*/
bool ComfIO::lockMutex(const char* where)
{
	if(!atomic)
	{
		return true;
	}
	if(!where)
	{
			where = "";
	}
	if(cons_debug)
	{
		writeConsole("Locking ", where);
	}
	if(::pthread_mutex_lock(&amutex) < 0)
	{
		stringstream s;
		s << "Mutex lock error: " << ::strerror(errno);
		throw IpcException(s.str(), __FILE__, __LINE__);
	}
	if(cons_debug)
	{
		writeConsole("Locked ", where);
	}
	return true;
}


/*!
	unlock mutex if atomicIO() is true

	Can throw:
						- IpcException
*/
bool ComfIO::unlockMutex(const char* where)
{
	if(!atomic)
	{
		return true;
	}
	if(!where)
	{
			where = "";
	}
	if(cons_debug)
	{
		writeConsole("Unlocking ", where);
	}
	if(::pthread_mutex_unlock(&amutex) < 0)
	{
		stringstream s;
		s << "Mutex unlock error: " << ::strerror(errno);
		throw IpcException(s.str(), __FILE__, __LINE__);
	}
	if(cons_debug)
	{
		writeConsole("Unlocked ", where);
	}
	return true;
}

/*! sets state of console debugging:\n
 			in:
				- true: 	console debugging will be turned on\n
				- false: 	console debugging will be turned off

	if set to true, every line will be displayed on the console\n
	 before it is written and after it was read
*/

void ComfIO::consoleDebug(bool on)
{
	cons_debug = on;
	if(on)
	{	
		ioctrls_chars 		 = new const char*[32]	;
		ioctrls_chars[0]	 = "NUL";
		ioctrls_chars[1]	 = "SOH";
		ioctrls_chars[2]	 = "STX";
		ioctrls_chars[3]	 = "ETX";
		ioctrls_chars[4]	 = "EOT";
		ioctrls_chars[5]	 = "ENQ";
		ioctrls_chars[6]	 = "ACK";
		ioctrls_chars[7]	 = "BEL";
		ioctrls_chars[8]	 = "BS";
		ioctrls_chars[9]	 = "HT";
		ioctrls_chars[10]	 = "LF";
		ioctrls_chars[11]	 = "VT";
		ioctrls_chars[12]	 = "FF";
		ioctrls_chars[13]	 = "CR";
		ioctrls_chars[14]	 = "SO";
		ioctrls_chars[15]	 = "SI";
		ioctrls_chars[16]	 = "DLE";
		ioctrls_chars[17]	 = "DC1";
		ioctrls_chars[18]	 = "DC2";
		ioctrls_chars[19]	 = "DC3";
		ioctrls_chars[20]	 = "DC4";
		ioctrls_chars[21]	 = "NAK";
		ioctrls_chars[22]	 = "SYN";
		ioctrls_chars[23]	 = "ETB";
		ioctrls_chars[24]	 = "CAN";
		ioctrls_chars[25]	 = "EM";
		ioctrls_chars[26]	 = "SUB";
		ioctrls_chars[27]	 = "ESC";
		ioctrls_chars[28]	 = "FS";
		ioctrls_chars[29]	 = "GS";
		ioctrls_chars[30]	 = "RS";
		ioctrls_chars[31]	 = "US";
	}
	else
	{
		if(ioctrls_chars)
		{
			delete[] ioctrls_chars;
		}
	}
}

/*!
	write str to stderr
*/


void ComfIO::writeConsole(const char* prefix, const char* str, size_t length)
{
	
	stringstream   s;
	struct timeval t;
	size_t len;
	gettimeofday(&t, 0);
	s << prefix << " at " << fixed << (t.tv_sec + t.tv_usec / 1000000.0) << " : ";
	cerr << s.str();
	if(length > 0)
	{
		len = length;
	}
	else
	{
		len = bytes_to_send ?  bytes_to_send : ::strlen(str);
	}
	for(unsigned int i = 0;i < len; i++)
	{
		unsigned short out = *(str + i) & 0x00ff;
		if(out > 127)
		{
			cerr << '<' << hex << showbase << out  << '>';

		}
		else if(::isprint(out))
		{
			if(bytes_to_send)
			{
				cerr << '<' << hex << showbase << out  << '>';
			}
			else
			{
				cerr << (char) out;
			}
		}
		else
		{
			if(out == 127)
			{
				cerr << "<DEL>";
			}
			else
			{
				cerr << '<' << ioctrls_chars[out] << '>';
			}
		}
	}
	cerr << dec << endl;
}

/*!
	clears internal buffer
*/

void ComfIO::clearBuffer()
{
	comfiostream.str("");
	::strcpy(comfiobuf, comfiostream.str().c_str());
	bytes_to_send = 0;
}

/*!
	checks whether input buffer is eots-terminated
	
	return:\n 
		- true: input buffer is eots-terminated\n
		- false: input buffer is not eots-terminated
*/

bool ComfIO::is_eots(const char* s)
{
	if(::strstr(s, eots->c_str()))
    {
        return true;
    }
	return false;
}

/*!
	checks whether input buffer is eots-terminated
	
	return:\n 
		- true: input buffer is eots-terminated\n
		- false: input buffer is not eots-terminated
*/

bool  ComfIO::is_eots(string& s)
{
	if(s.find(*eots) != string::npos)
	{
		return true;
	}
	return false;
}

/*!
 
					set the end of transmission string, i.e. the sequence of characters
					that will cause the callback function to execute if it is set
					
					in: 
						- eot:		end of transmission string 	
 	
*/

void ComfIO::setEOTS(const char* eot)
{
	*eots = eot; 	
}


/*!
 
					set the end of transmission string, i.e. the sequence of characters
					that will cause the callback function to execute if it is set
					
					in: 
						- eot:		end of transmission string 	
 	
*/

void ComfIO::setEOTS(string& eot)
{
	*eots = eot;
}

/*!
 
					get the end of transmission string, i.e. the sequence of characters
					that will cause the callback function to execute if it is set
					
					retunr: pointer to end of transmission string 	
 	
*/

const char *ComfIO::getEOTS()
{
	return eots->c_str();
}


/*! 
					sets the size of the input buffer to newsize

					any content already existing in the buffer will be kept\n

					the method can fail if:
					- newsize < old size and the existing content won't fit into the resized buffer\n
					- on memory allocation failure

 					in - 
						newsize:    the new size of the buffer

					Can throw:

						- IpcException


*/


bool ComfIO::setIOBufSize(ssize_t newsize)
{
	char *tmpptr = comfiobuf;
	char *newbuf;

	if( (long) ::strlen(tmpptr) > newsize)
	{
		throw IpcException("New buffer too small", __FILE__, __LINE__);
		return false;
	}
	if(!(newbuf = new char[newsize]))
	{
		throw IpcException("Mem allocation for new buffer failed", __FILE__, __LINE__);
		return false;
	}
	if(::strlen(tmpptr) > 0)
	{
		::strcpy(newbuf, tmpptr);
	}
	comfiobuf = newbuf;
	delete tmpptr;
	max_msg = newsize;
	return true;
}

/*!
	private method, calls the callback function if any is set
//					and the input buffer ends with the eots string

*/

ssize_t ComfIO::has_eots()
{
	if(eots_call_ptr && is_eots(comfiostream.str().c_str()))
	{
		return (*eots_call_ptr)();
	}
	return comfiostream.str().size();
}


/*!
	returns the current maximum size of the input buffer
*/

ssize_t ComfIO::getIOBufSize()
{
	return max_msg;
}

/*! 				writes data to the input buffer\n
					since this method is to be used to
					write binary data to the buffer, it will
					not execute the callback function

					in - buf:		pointer to data
					
					in - nbytes:	length of data buffer

					returns:		 
					- number of bytes written to the input buffer

				Can throw:

				- IpcException
*/

ssize_t ComfIO::write(const void *buf, size_t nbytes)
{
    unsigned char *ptr = (unsigned char*) buf;      // Compilersedativ
	size_t i;
	
	if(nbytes > SSIZE_MAX)
	{
		throw IpcException("Large input buffer", __FILE__, __LINE__);
	}

	if(nbytes > comfiostream.str().max_size())
	{
			throw IpcException("Large input buffer", __FILE__, __LINE__);
	}
	for(i = bytes_to_send = 0; i < nbytes; i++)
	{
		comfiostream << ptr[i];
		++bytes_to_send;
	}
	return comfiostream.str().size();
}

/*!
				   	writes a formatted string to the input buffer

		in: 
			- format:	the format string (cf. man 3 printf)

		in: 
			- ap:	an initialized argument list, i.e. va_start has been
					called for ap and va_end should be called for ap
					after this method returns.

		returns:   
		
					- if no callback function is set or the string is not eots-terminated:	
 					the number of bytes written to the output buffer or
					-1 on error

					- if a callback function is set and the string is eots-terminated:
					the return value of the callback function

					This method is meant to be used in derived classes
					where you want to implement your own method(const char* format, ...)
		Can throw:

				- IpcException
		
*/


ssize_t ComfIO::fwriten(const char* format, va_list ap)
{
	ssize_t retval;
	retval = ::vsnprintf(comfiobuf, max_msg, format, ap);
	if(retval > max_msg)
	{
		throw IpcException("Large input buffer", __FILE__, __LINE__);
	}
	comfiostream << comfiobuf;
	return has_eots();
}

/*!
		split string into tokens and put them into out

		in: - in:	the string to split\n
		in: - out:	the vector the tokens will be stored in\n
		in: - zero: if true, the vector will be set to zero size before putting in tokens
		in:	- delims: the characters considered as token separators\n
					  (if delim is 0, " \t\n\r\f\v" will be taken as separators, cf. man isspace)\n

		returns:	the number of tokens found in the string
*/   

ssize_t ComfIO::tokenize(string& in, vector<string>& out, bool zero, const char* delims)
{
	if(in.size() == 0)
	{
		return 0;
	}
	if(zero)
	{
		// set vector to zero lenght
		out.resize(0);
	}
	if(!delims)
	{
		delims = " \t\n\r\f\v";
	}
	size_t start, end;
	// discard any leading delimiters
	start = in.find_first_not_of(delims);
	// now loop throug the string
	while((end = in.find_first_of(delims, start)) != string::npos)
	{
		out.push_back(in.substr(start, end - start));
		start = in.find_first_not_of(delims, end);
	}
	// if there is something left after the last delimiter, 
	// but the string does not terminate with a delimiter
	// be nice and put it into the vector
	if(end == string::npos && start < in.size())
	{
		out.push_back(in.substr(start));
	}	
	return out.size();
}

/*!
				   	writes a formatted string to the input buffer

		in: - format:	the format string (cf. man 3 printf)

		in: - ...:		variable list of arguments	
					
					

		returns:   
				
					- if no callback function is set or the string is not eots-terminated:	
 					the number of bytes written to the output buffer or
					-1 on error

					- if a callback function is set and the string is eots-terminated:
					the return value of the callback function

		

*/



ssize_t ComfIO::printf(const char* format, ...)
{
	va_list ap;
	ssize_t retval;
	::va_start(ap, format);
	retval = ::vsnprintf(comfiobuf, max_msg, format, ap);
	::va_end(ap);
	if(retval > max_msg)
	{
		throw IpcException("Large input buffer", __FILE__, __LINE__);
	}
	comfiostream << comfiobuf;
	return has_eots();
}

ssize_t	ComfIO::vscanf(const char *format, va_list ap)
{
	int retval = ::vsscanf(comfiobuf, format, ap);
	return retval;
}

/*!
	return: the address of the TSpecificFunctor or NULL, if none is set
*/


TFunctor* ComfIO::getActOnEOTS()
{	
	return eots_call_ptr;
}

/*!
	set a  TSpecificFunctor callback function which will be executed
	if eots is written to the input buffer

	in - ptr: address of TSpecificFunctor callback function or NULL
	to delete the existing  TSpecificFunctor
*/ 

void  ComfIO::setActOnEOTS(TFunctor* ptr)
{
	if(!ptr && eots_call_ptr)
	{
		delete eots_call_ptr;
	}
	else
	{
		eots_call_ptr = ptr;
	}
}

/*!
	write c to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(char c)
{
	comfiostream << c;
	has_eots();
    return *this;
}

/*!
	write uc to the input buffer

	return: reference to object
	
*/
ComfIO& ComfIO::operator<<(unsigned char uc)
{
	comfiostream << uc;
	has_eots(); 
	return *this;
}


/*!
	write sc to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(signed char sc)
{
	comfiostream << sc; 
	has_eots(); 
	return *this;
}


/*!
	write *ucptr to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(const unsigned char *ucptr) 
{
	comfiostream << ucptr;
	has_eots(); 
	return *this;
}


/*!
	write *scptr to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(const signed char *scptr) 
{
	comfiostream << scptr;
	has_eots(); 
	return *this;
}


/*!
	write *cptr to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(const char *cptr)
{
	comfiostream << cptr;
	has_eots();
    return *this;
}


/*!
	write str to the input buffer

	return: reference to object
	
*/

ComfIO&  ComfIO::operator<<(string& str)
{
	comfiostream << str;
	has_eots(); 
	return *this;
}

/*!
	write content of str to the input buffer

	(nothing will be written if str is NULL)

	return: reference to object
	
*/


ComfIO& ComfIO::operator<<(string* str)
{
	if(str)
	{
		comfiostream << *str;
	}
	has_eots();
	return *this;
}

/*!
	convertt s to ascii and write it to the input buffer
	
	return: reference to object	
*/

ComfIO& ComfIO::operator<<(short s)
{
	comfiostream << s;
	has_eots();
    return *this;
}


/*!
	convertt su to ascii and write it to the input buffer
	
	return: reference to object	
*/


ComfIO& ComfIO::operator<<(unsigned short su)
{
	comfiostream << su;
	has_eots();
    return *this;
}


/*!
	convertt i to ascii and write it to the input buffer
	
	return: reference to object	
*/


ComfIO& ComfIO::operator<<(int i)
{
	comfiostream << i;
	has_eots();
    return *this;
}


/*!
	convertt ui to ascii and write it to the input buffer
	
	return: reference to object	
*/


ComfIO& ComfIO::operator<<(unsigned int ui)
{
	comfiostream << ui;
	has_eots();
    return *this;
}


/*!
	convertt l to ascii and write it to the input buffer
	
	return: reference to object	
*/


ComfIO& ComfIO::operator<<(long l)
{
	comfiostream << l;
	has_eots();
    return *this;
}

/*!
	convertt ul to ascii and write it to the input buffer
	
	return: reference to object	
*/

ComfIO& ComfIO::operator<<(unsigned long ul)
{
	comfiostream << ul;
	has_eots();
    return *this;
}

/*!
	convertt f to ascii and write it to the input buffer
	
	return: reference to object	
*/

ComfIO& ComfIO::operator<<(float f)
{
	comfiostream << f;
    return *this;
}

/*!
	convertt d to ascii and write it to the input buffer
	
	return: reference to object	
*/

ComfIO& ComfIO::operator<<(double d)
{
	comfiostream << d;
    return *this;
}

/*!
	convertt ld to ascii and write it to the input buffer
	
	return: reference to object	
*/
ComfIO& ComfIO::operator<<(long double ld)
{
	comfiostream << ld;
    return *this;
}


/*!
	 write output manipulator pf to the input buffer

	return: reference to object
	
*/


ComfIO& ComfIO::operator<<(ostream& ( *pf )(ostream&))
{
	comfiostream << pf;
	has_eots();
	return *this;
}


/*!
	 write format flag pf to the input buffer

	return: reference to object
	
*/


ComfIO& ComfIO::operator<<(ios_base& ( *pf )(ios_base&))
{
	comfiostream << pf;
	return *this;
}

/*!
	 write format flag pf to the input buffer

	return: reference to object
	
*/

ComfIO& ComfIO::operator<<(ios& ( *pf )(ios&))
{
	comfiostream << pf;
	return *this;
}

/*ComfIO& ComfIO::operator>>(string& str)
{

}

ComfIO& ComfIO::operator>>(int i)
{

}


ComfIO& ComfIO::operator>>(double d)
{

}*/

/*!
	 convert cptr to double and return it

	Can throw:

	- IpcConvertException
	
*/

double  ComfIO::toDouble(const char *cptr)
{
	char* endptr;
	double d = strtod(cptr, &endptr);
	if(*endptr)
	{
		throw IpcConvertException("Double convertsion error", __FILE__, __LINE__, cptr);
	}
	return d;
}

/*!
	 convert cptr to float and return it

	Can throw:

	- IpcConvertException
	
*/

float  ComfIO::toFloat(const char *cptr)
{
	char* endptr;
	float f = strtof(cptr, &endptr);
	if(*endptr)
	{
		throw IpcConvertException("Float convertsion error", __FILE__, __LINE__, cptr);
	}
	return f;
}

/*!
	 convert cptr to unsigned long using base and return it

	Can throw:

	- IpcConvertException
	
*/

unsigned long ComfIO::toUlong(const char *cptr, int base)
{
	checkBase(base, cptr);
	char* endptr;
	unsigned long ul = strtoul(cptr, &endptr, base);
	if(*endptr)
	{
		throw IpcConvertException("ulong convertsion error", __FILE__, __LINE__, cptr);
	}
	return ul;

}

/*!
	 convert cptr to unsigned long long using base and return it

	Can throw:

	- IpcConvertException
	
*/

unsigned long long	ComfIO::toUlongLong(const char *cptr, int base)
{
	checkBase(base, cptr);
	char* endptr;
	unsigned long long ull = strtoull(cptr, &endptr, base);
	if(*endptr)
	{
		throw IpcConvertException("long long convertsion error", __FILE__, __LINE__, cptr);
	}
	return ull;
}
/*!
	 convert cptr to long using base and return it

	Can throw:

	- IpcConvertException
	
*/


long ComfIO::toLong(const char *cptr, int base)
{
	checkBase(base, cptr);
	char* endptr;
	unsigned long l = strtol(cptr, &endptr, base);
	if(*endptr)
	{
		throw IpcConvertException("long convertsion error", __FILE__, __LINE__, cptr);
	}
	return l;

}
/*!
	 convert cptr to long long using base and return it

	Can throw:

	- IpcConvertException
	
*/

long long ComfIO::toLongLong(const char *cptr, int base)
{
	checkBase(base, cptr);
	char* endptr;
	long long ll = strtoll(cptr, &endptr, base);
	if(*endptr)
	{
		throw IpcConvertException("long long convertsion error", __FILE__, __LINE__, cptr);
	}
	return ll;
}

void ComfIO::checkBase(int base, const char* cptr)
{
	if(base < 0 || base > 36)
	{
		stringstream tmp;
		tmp << "Invalid base: " << base;
		throw IpcConvertException(tmp.str().c_str(),  __FILE__, __LINE__, cptr);
	}
}
