#include "io.h"
#include "ipc_exception.h"

#include <iostream>

extern "C"
{
	#include <termios.h>
}

/*!
 *
 					in:

						- client : connection established with TcpSocket

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered

						- destroy_client : if true, destroy client after we got the file descriptor

 
					constructor for Io using connection established with TcpSocket

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true

*/

Io::Io(TcpSocket* client, const char* eots, bool send_on_eots, bool destroy_client) : ComfIO(1024) 
{
	fd = client->filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
	if(destroy_client)
	{
		delete client;
	}
}


/*!
 *
 					in:

						- client : connection established with TcpSocket

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered


 
					constructor for Io using connection established with TcpSocket

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true

*/

Io::Io(TcpSocket& client, const char* eots, bool send_on_eots)
{
	fd = client.filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
}

/*!
 *
 					in:

						- client : connection established with DomainSocket

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered

						- destroy_client : if true, destroy client after we got the file descriptor

 
					constructor for Io using connection established with TcpSocket

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true

*/

Io::Io(DomainSocket* client, const char* eots, bool send_on_eots, bool destroy_client) : ComfIO(1024) 
{
	fd = client->filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
	if(destroy_client)
	{
		delete client;
	}
}

/*!
 *
 					in:

						- client : connection established with DomainSocket

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered


 
					constructor for Io using connection established with TcpSocket

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true

*/

Io::Io(DomainSocket& client, const char* eots, bool send_on_eots)
{
	fd = client.filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
}

/*!
 					in:

						- client : connection established with TtyDevice

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered

						- destroy_client : if true, destroy client after we got the file descriptor


 
					constructor for Io with connection established to TtyDevice

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true


*/

Io::Io(TtyDevice* client, const char* eots, bool send_on_eots, bool destroy_client) : ComfIO(1024)
{
	fd = client->filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
	if(destroy_client)
	{
		delete client;
	}
}

/*!
 					in:

						- client : connection established with TtyDevice

						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered



 
					constructor for Io with connection established to TtyDevice

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to true

					sets read_select and write_select to true

					sets close_on_delete to true


*/

Io::Io(TtyDevice& client, const char* eots, bool send_on_eots)
{
	fd = client.filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
}

/*!
					in:


						- eots   : end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered


 
					constructor for Io with no connection established

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

					sets glErr.getEOTS() (= end of transmission string) to 0xa

 					sets send_at_once to false

					sets read_select and write_select to true

					sets close_on_delete to false


					This Io will not be usable before a connection was set with Io::client(TcpSocket&)\n
					or Io::client(TtyDevice& client)
*/

Io::Io(const char* eots, bool send_on_eots) : ComfIO(1024)
{
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(false);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
	
}


Io::Io(USBDevice* client, const char* eots, bool send_on_eots, bool destroy_client) : ComfIO(1024)
{
	fd = client->filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
	if(destroy_client)
	{
		delete client;
	}
}

Io::Io(USBDevice& client, const char* eots, bool send_on_eots)
{
	fd = client.filedescriptor();
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(true);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
}

/*!
					in:

						- filedes		: (valid, i.e. opened file descriptor)

						- eots   		: end of transmission string

						- send_on_eots : if true, send operation will be called if eots is encountered


 
					constructor for Io with file descriptor filedes

					assigns default value 2 secs for select timeout

					allocates io buffer memory with 1024 byte

 					sets send_at_once to false

					sets read_select and write_select to true

					sets close_on_delete to false


*/

Io::Io(int filedes, const char* eots, bool send_on_eots) : fd(filedes)
{
	if(fd < 0)
	{
		throw IpcException("Invalid file descriptor", __FILE__, __LINE__);	
	}
	
	writeTimeout(2, 0);
	readTimeout(2, 0);
	closeOnDelete(false);
	readSelect(true);
	writeSelect(true);
	setEOTS(eots);
	sendOnEOTS(send_on_eots);
}

/*!
	set connection to use to the one established in TcpSocket client
*/

bool Io::client(TcpSocket& client)
{
	fd = client.filedescriptor();
	return true;
	
}

/*!
	set connection to use to the one established in TtyDevice client
*/

bool Io::client(TtyDevice& client)
{
	fd = client.filedescriptor();
	return true;
}

/*!
	set connection to use to the one established in DomainSocket client
*/

bool Io::client(DomainSocket& client)
{
	fd = client.filedescriptor();
	return true;
}
	

/*!
	Io destructor, closes connection if Io::closeOnDelete() was set to true
*/

Io::~Io()
{
	if(close_on_delete)
	{
		::close(fd);
	}
}

//+----------------------------------------------------------------------------
//
// 
// check whether filedes is ready for reading\n
//
// return value: 1, if file decriptor is ready
//
// Can throw:
//
//
// - IpcException
//-----------------------------------------------------------------------------


int	Io::select_read()
{
	int retval;
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	if( (retval = ::pselect(fd + 1, &fds, NULL, NULL, &rts, NULL)) < 1)
	{
		if(retval == 0)
		{
			throw IpcTimeOutException("read timeout", __FILE__, __LINE__);
		}
		else
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		return retval;
	}
	return retval;
}

//+----------------------------------------------------------------------------
//
// method : 		Io::select_write()
// 
// description :   	check whether filedes is ready for writing
// returns:		    1, if file decriptor is ready
//
// Can throw:
//
//
// - IpcException
//-----------------------------------------------------------------------------

int	Io::select_write()
{
	int retval;
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	if( (retval = ::pselect(fd + 1, NULL, &fds, NULL, &wts, NULL)) < 1)
	{
		if(retval == 0)
		{
			throw IpcTimeOutException("write timeout", __FILE__, __LINE__);
		}
		else
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		return retval;
	}
	return retval;
}
/*!
 
		sets timeout values for writing to socket

 		in : 
			- sec: 		seconds of timeout

		in : 
			- nsec:		nanoseconds of time
*/

void Io::writeTimeout(long sec, long nsec)
{
	wts.tv_sec 	= sec;
	wts.tv_nsec 	= nsec; 
}

/*!
 
		sets timeout values for reading from  socket

 		in :
			- sec: 		seconds of timeout

		in :
			-nsec:		nanoseconds of time
*/

void Io::readTimeout(long sec, long nsec)
{
	rts.tv_sec 	= sec;
	rts.tv_nsec 	= nsec; 
}

/*!
	gets timeout values for writing to socket

 	out: 
		- sec: 		seconds of timeout

 	out:
		- nsec:		nanoseconds of time
*/

void  Io::writeTimeout(long *sec, long *nsec)
{
	*sec = wts.tv_sec;
	*nsec = wts.tv_nsec;
}

/*!
	gets timeout values for reading from socket

 	out: 
		- sec: 		seconds of timeout

 	out: 
		- nsec:		nanoseconds of time
*/

void  Io::readTimeout(long *sec, long *nsec)
{
	*sec = rts.tv_sec;
	*nsec = rts.tv_nsec;
}


/*!
		calls ComfIO::setActOnEOTS() setting Io::Send() as callback function, i.e.
		all buffer data will be written to the socket immediately when eots is written to
		the buffer 	

	in - b: 

		- true: data will be written to socket if eots is written to output buffer

		- false:  data will not be written to socket if eots is written to output buffer
*/

void Io::sendOnEOTS(bool b)
{
	if(b)
	{
		setActOnEOTS(new TSpecificFunctor<Io>(this, &Io::Send));
	}
	else
	{
		setActOnEOTS(NULL);
	}		
}

/*!
				   	writes output buffer data to the socket and clears
					output buffer after having written all data.

					return: number of bytes written to the socket

	 Can throw:


 	- IpcException


*/

ssize_t Io::Send()
{
	ssize_t nwritten = 0;
	string tmpstr 	= buffer()->str();
	string send_str = buffer()->str();
	ssize_t nleft, nbytes;
	//nleft =  nbytes = tmpstr.size();
	
	if(getBytesToSend()) // binary data
	{
		nleft =  nbytes = getBytesToSend();
	}
	else
	{
		nleft =  nbytes = tmpstr.size();
	}
	if(store_b)
	{
		store_buffer = tmpstr;
	}
	while (nleft > 0)
	{
		if(write_select)
		{
			select_write();
		}
		nwritten = ::write(fd, send_str.c_str(), nleft);
		if (nwritten <= 0)
        {
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
			
        }
		if( (nleft -= nwritten) > 0)
		{
			send_str = tmpstr.substr(nwritten, nleft);
		}
	}
	if(consoleDebug() && nwritten > 0)
	{
		writeConsole("Send", send_str.c_str());
	}
	clearBuffer();
	return(nbytes - nleft);
}

/*!
				   	send binary data

		in: 
			- buf:		pointer to data to send

		in: 
			- nbytes:	number of bytes to send

		returns: 

		- number of bytes send

		Can throw:

		- IpcException


*/

ssize_t	Io::SendBin(const void *buf, size_t nbytes)
{
	if(write(buf, nbytes) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	return Send();
}

/*!
				   	writes vector data to the file descriptor

					return: number of bytes written to the file descriptor

	 Can throw:


 	- IpcException


*/

ssize_t Io::writen(vector<unsigned char>& buf)
{
	std::vector<unsigned char>::iterator it;
	ssize_t nwritten;
	size_t retval = 0;
	for(;;)
	{
		if(write_select)
		{
			select_write();
		}
		nwritten = ::write(fd, buf.data() + retval, buf.size() - retval);
		if (nwritten < 0)
        {
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
        }
		retval += nwritten;
		if(buf.size() == retval)
		{
			break;
		}
	}
	return retval;

}


/*! 		  
					reads up to nbytes bytes from the file descriptor

					in: 
						- buf:		buffer to write the data to

					in: 
						- nbytes:   max. number of bytes to read
					in:
						- readall: if true, the method will only return if\n
								   either nbytes were read or an exception is thrown\n
								   if false, the method will behave like ::read, i.e. it\n
                                   is not an error if less than nbytes were read

					returns:
					
						- number of bytes read

	Can throw:

					- IpcException


*/



ssize_t Io::readn(const void *buf, size_t nbytes, bool readall)
{
	ssize_t  nleft = nbytes, nread = 0;
    char* ptr = (char*) buf;
	
	 
   
	if(nbytes > SSIZE_MAX)
	{
		throw IpcException( "Large buffer size", __FILE__, __LINE__);
	}
	while (nleft > 0)
    {
		if(read_select)
		{
			select_read();
		}
        nread = ::read(fd, ptr, nleft);
        //cerr << "Read: " << nread << endl;
        if (nread < 0)
        {
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
        }
        else if (!nread)
		{
			if(readall)
			{
				continue;
			}
			else
			{
            	break;
			}
		}
		else
		{
        	nleft -= nread;
        	ptr   += nread;
			if(readall)
			{
				continue;
			}
			else
			{
				break;
			}
		}
    }
	if(consoleDebug() && nread > 0)
	{
		writeConsole("Received", (const char*) buf, nbytes - nleft);
	}
    return(nbytes - nleft);
}

/*! 		  
					reads up to nbytes bytes from the file descriptor

					in: 
						- buf:		vector to write the data to

					in: 
						- nbytes:   max. number of bytes to read
					in:
						- readall: if true, the method will only return if\n
								   either nbytes were read or an exception is thrown\n
								   if false, the method will behave like ::read, i.e. it\n
                                   is not an error if less than nbytes were read

					returns:
					
						- number of bytes read

	Can throw:

					- IpcException


*/


ssize_t	Io::readn(vector<unsigned char>& buf, ssize_t nbytes, bool readall)
{
	unsigned char c;
	ssize_t nread, retval = 0;

	for(;;)
    {
		if(read_select)
		{
			select_read();
		}
		nread = ::read(fd, &c, 1);
		if(nread < 0)
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		if(nread == 0)
		{
			if(readall)
			{
				continue;
			}
			else
			{
				break;
			}
		}
		if(nread == 1)
		{
			buf.push_back(c);
			++retval;
		}
		if(retval == nbytes)
		{
			break;
		}
		
	}
	if(consoleDebug() && nread > 0)
	{
		writeConsole("Received", (const char*) buf.data(), nread);
	}

	return buf.size();
}


/*! 
				  	reads from the connection until either maxlen data where read
					or an eots was encountered 

 					in: 
						- buf:			buffer to write the data to

					in: 
						- nbytes:		length of buf ( <= SSIZE_MAX)

					returns:
			 		
						- number of bytes read

					Can throw:

					- IpcException


*/

ssize_t Io::readLine(const void *buf, size_t maxlen)
{

    ssize_t n, rc;
    char c; 
    char *cptr = (char*) buf;
	::memset(cptr, 0, maxlen);
	if(maxlen > SSIZE_MAX)
	{
		throw IpcException("IO buffer too big", __FILE__, __LINE__);
	}
    for (n = 0; n < (ssize_t) maxlen; n++) 
    {
		if(read_select)
		{
			select_read();
		}
    	if ( (rc = ::read(fd, &c, 1)) == 1 )
		{
			*cptr = c;
			++cptr;
			if(is_eots((const char*) buf))
			{
				if(consoleDebug())
				{
					writeConsole("Received", (const char*) buf);
				}
				*(::strstr((char *) buf, getEOTS())) = 0;
				n = ::strlen((char *) buf);
				break;
			}
    	}
    	else if (rc == 0) // EOF
    	{
			n = rc;
			break;
		}
    	else
    	{ 
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
    	}
	}
    return n;
}

 
/*!					reads from the connection until either str.max_size() data where read
					or an eots was encountered 

				in: 
					- str:		string object to write data to
	
				returns: 		
					- number of bytes read

				Can throw:

					- IpcException


*/


ssize_t Io::readLine(string& str)
{

    ssize_t rc; 
    size_t 	n;
	char 	c;
	str.clear();
    for (n = 0; n < str.max_size(); n++) 
    {
		if(read_select)
		{
			select_read();
		}
    	if ((rc = ::read(fd, &c, 1)) == 1)
        {
            //cerr << "Received: " << hex << (int) c << endl;
			str += c;
			if(is_eots(str))
			{
				if(consoleDebug())
				{
					writeConsole("Received", str.c_str());
				}
				str.erase(str.rfind(getEOTS()));
				n = str.size();
				break;
			}
        }
        else if (rc == 0) // EOF
        {
			n = rc;
			break;
		}
		else
        {
            throw IpcException(::strerror(errno), __FILE__, __LINE__);
        }
    }
	if(n > SSIZE_MAX)
	{
		throw IpcException("Read too much", __FILE__, __LINE__);
	}
    return n;
}

/*!
				   	read a formatted string from Io

					in: 
						- format:	the format string (cf. man 3 scanf)

						- ... adresses of variables to take up values
				returns: 

						- the number of successfull format conversions

				Can throw:

					- IpcException


					- IpcScanfException
		
*/

ssize_t	Io::scanf(const char *format, ...)
{
	int conv;
	const char *c;
	// find number of wanted conversions
	for(conv = 0, c = format; *c; c++)
	{
		if(*c == '%' && *(c+1) == '%')
		{
			++c;
		}
		else if(*c == '%' && *(c+1) != '%')
		{
			++conv;
		}
	}
	readLine(comfiobuf, max_msg);
	va_list ap;
	ssize_t retval;
	::va_start(ap, format);
	retval = ComfIO::vscanf(format, ap);
	::va_end(ap);
	if(retval != conv)
	{
		throw IpcScanfException("Mismatched input/format strings", __FILE__, __LINE__, format, comfiobuf);
	}
	return retval;
}


/*!
		read line, split it into tokens and put them into out

		in: 
			- out:	the vector the tokens will be stored in\n
		in: 
			- zero: if, true, the vector will be resized to zero before putting tokens in (default is true)\n 
		in:	
			- delims: the characters considered as token separators\n
					  (if delim is 0, " \t\n\r\f\v" will be taken as separators)\n

		returns:
			
		- the number of tokens, i.e. the size of out

		Can throw:

		- IpcException


*/

ssize_t Io::tokenize(vector<string>& out,  bool zero, const char* delims)
{
	string in;
	ssize_t retval;
	if( (retval = readLine(in)) < 1)
	{
		return (size_t) retval;
	}
	retval = ComfIO::tokenize(in, out, zero, delims);
	return retval;
}

void Io::flushRead()
{
	if(is_serial())
	{
		if(::tcflush(fd, TCIFLUSH) < 0)
		{
			throw IpcException(::strerror(errno),  __FILE__, __LINE__);
		}
	}
}


void Io::flushWrite()
{
	if(is_serial())
	{
		if(::tcflush(fd, TCOFLUSH) < 0)
		{
			throw IpcException(::strerror(errno),  __FILE__, __LINE__);
		}
	}
}

void Io::flushAll()
{
	if(is_serial())
	{
		if(::tcflush(fd, TCIOFLUSH) < 0)
		{
			throw IpcException(::strerror(errno),  __FILE__, __LINE__);
		}
	}
}

bool Io::is_serial()
{
	struct stat buf;
	if(fstat(fd, &buf) < 0)
	{
		throw IpcException(::strerror(errno),  __FILE__, __LINE__);
	}
	if(S_ISCHR(buf.st_mode))
	{
		return true;
	}
	return false;
}

