#ifndef IPC_IO_H
#define IPC_IO_H


#include "comfio.h"
#include "ipc_tty.h"
#include "ipc_tcp.h"
#include "ipc_usbdevice.h"
/*!

	Class represents a usable io connection

	This is the working class containing most of the functionality
 
*/

class Io : public ComfIO
{
	private:
		struct 			timespec rts;
		struct 			timespec wts;
		bool			close_on_delete;
		bool			store_b;
		string			store_buffer;
		bool 			is_serial();
			

	protected:
		int 			fd;
		int				select_read();
		int				select_write();
		bool			read_select;
		bool			write_select;
  	public:
		Io(TcpSocket* client, const char* eots = "\n", bool send_on_eots = true, bool destroy_client = true);
		Io(TcpSocket& client, const char* eots = "\n", bool send_on_eots = true);
		Io(TtyDevice* client, const char* eots = "\n", bool send_on_eots = true, bool destroy_client = true);
		Io(TtyDevice& client, const char* eots = "\n", bool send_on_eots = true);
		Io(USBDevice* client, const char* eots = "\n", bool send_on_eots = true, bool destroy_client = true);
		Io(USBDevice&, const char* eots = "\n", bool send_on_eots = true);
		Io(DomainSocket* client, const char* eots = "\n", bool send_on_eots = true, bool destroy_client = true);
		Io(DomainSocket& client, const char* eots = "\n", bool send_on_eots = true);
		Io(const char* eots = "\n", bool send_on_eots = true);
		Io(int filedes, const char* eots = "\n", bool send_on_eots = true);
		~Io();
		bool		client(TcpSocket& client);
		bool		client(TtyDevice& client);
		bool		client(DomainSocket& client);
		/*!
			if called with true, the filesdescriptor will be closed on deletion
		*/
		void		closeOnDelete(bool b)	{close_on_delete = b;}
		/*!
			return: the current state of close_on_delete flag
		*/
		bool		closeOnDelete()			{return close_on_delete;}
		void 		writeTimeout(long sec, long nsec);
		void 		readTimeout(long sec, long nsec);
		void 		writeTimeout(long *sec, long* nsec);
		void 		readTimeout(long *sec, long* nsec);
		/*!
			set read_select flag

			in  
				- true: pselect will be called before socket read operations

				- false: pselect will not be called before socket read operations
		*/
		void		readSelect(bool b)		{read_select = b;}
		/*!
		set write_select flag

		in - b: 
		- true: pselect will be called before socket write operations 
		- false: pselect will not be called before socket write operations
		*/
		void		writeSelect(bool b)		{write_select = b;}
		/*!
			return: the current state of read_select flag
		*/
		bool		readSelect()			{return read_select;}
		/*!
			return: the current state of write_select flag
		*/
		bool		writeSelect()			{return write_select;}
		void		setEOTS(const char* eot){ComfIO::setEOTS(eot);}
		void		setEOTS(string& eot)    {ComfIO::setEOTS(eot);}
		/*!
			return

			- true: a callback function is set

			- false: no callback function is set
		*/
		bool		sendOnEOTS()			{return (bool) getActOnEOTS();}
		void		sendOnEOTS(bool b);
		/*!
			 if set to true, the internal buffer is stored after writing\n
			 and can be read with Io::storeBuffer()
		*/
		void		store(bool b)			{store_b = b;}
		/*!
 			returns state of buffer storing
		*/
		bool		store()					{return store_b;}
		/*! 
 			returns io buffer stored after sending (if Io::store was called with true)
		*/
		string&		storeBuffer()			{return store_buffer;}
		ssize_t  	Send();
		ssize_t		SendBin(const void *buf, size_t nbytes);
		ssize_t		writen(vector<unsigned char>& buf);
		/*!
			write binary data
		
			in:
				- buf: buffer with data

				- nbytes: amount of bytes to write

			return: 
	
				number of bytes writtenn

			may throw IpcException
		*/
		ssize_t		writen(const void *buf, size_t nbytes) {return SendBin(buf, nbytes);}
		ssize_t  	readn(const void *buf, size_t nbytes, bool readall = false);
		ssize_t  	readn(vector<unsigned char>& buf, ssize_t nbytes, bool readall = false);
		ssize_t  	readLine(const void *buf, size_t maxlen);
		ssize_t  	readLine(string& str);
		/*!			reads from the connection until either str.max_size() data where read
					or an eots was encountered 

				
			in: 
					- str:		string object to write data to
	
			returns: 		
					- number of bytes read

		Can throw:

			IpcException

			IpcTimeoutException

*/

		ssize_t  	readLine(string* str) {return readLine(*str);}
		ssize_t		scanf(const char *format, ...);
		ssize_t		tokenize(vector<string>& out, bool zero = true, const char* delims = 0);
		/*!
			flush read buffer of serial line, does nothing on socket
		*/
		void flushRead();
		/*!
			flush write buffer of serial line
		*/
		void flushWrite();
		/*!
			flush read/write buffer of serial line
		*/
		void flushAll();


};

#endif



