#ifndef IOEXCP_H
#define IOEXCP_H

#include <exception>
#include <string>

using namespace std;

#define IPC_PRINTEX print(__FILE__, __LINE__)


/*!
	IpcException class

	Base exception class. This type of exception may be thrown by nearly all constructors

	and method calls

*/

class IpcException : public exception
{
	public:
		IpcException()								  		{}
		/*!
 					constructor for IpcException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcException(const char* reas, const char* fn, int ln)
			: reason(reas), file(fn), line(ln) {}
		/*!
 					constructor for IpcException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcException(string reas, string fn, int ln)
			: reason(reas), file(fn), line(ln) {}
		/*!
 					constructor for IpcException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcException(const char* reas, string fn, int ln)
			: reason(reas), file(fn), line(ln) {}
		/*!
 					constructor for IpcException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcException(string reas, const char* fn, int ln)
			: reason(reas), file(fn), line(ln) {}
		/*!
 					destructor for IpcException
		*/
		~IpcException() throw()						  		{}
		/*!
				returns exception description
		*/
		virtual	const char*	what() const throw() {return reason.c_str();}
		/*!
				returns line number in source file
		*/
		virtual int			lineNumber() const throw() {return line;}
		/*!
				returns source file name
		*/
		virtual const char*	fileName() const throw()  {return file.c_str();}
		virtual void 		print() const throw();
		virtual void 		print(const char* fn, int ln) const throw();
		/*! 
 			enum with constants for the exception type:
			GENERAL
			SCANF
			TIMEOUT
			CONVERT
				
		*/
		enum excptype {GENERAL, SCANF, TIMEOUT, CONVERT};
		/*! 
 			returns classname
		*/
		virtual const char*	classname() const throw() {return "IpcException";}
		/*! 
 			returns exception type as declared in excptype
		*/
		virtual int			type() {return GENERAL;}
	protected:
		string reason;
		string file;
		int	   line;
};

/*!
	IpcScanfException class


	scanf exception class. This type of exception may be thrown by Io::scanf()

*/



class IpcScanfException : public IpcException
{
	public:
		IpcScanfException() {}
		/*!
 					destructor for IpcScanfException
		*/
		~IpcScanfException ()throw() {}
		/*!
 					constructor for IpcScanfException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

					in : 
						- fm   : scanf format string

					in : 
						- in 	: scanf input string

		*/
		IpcScanfException(const char* reas, const char* fn, int ln, const char* fm, const char* in)
			: IpcException(reas, fn, ln), format(fm), input(in) {}
		/*!
				returns format string of failed scanf call
		*/
		virtual const char*	formatString() const throw() {return format.c_str();}
		/*!
				returns input string of failed scanf call
		*/
		virtual const char*	inputString() const throw() {return input.c_str();}
		virtual void 		print() const throw();
		virtual void		print(const char* fn, int ln) const throw();
		/*! 
 			returns class name
		*/
		virtual const char*	classname() const throw() {return "IpcScanfException";}
		virtual int			type() {return SCANF;}
	private:
		string format, input;
};

/*!
	IpcTimeOutException class


	timeout exception class. This type of exception may be thrown  on connect /read / write timeouts

*/

class IpcTimeOutException : public IpcException
{
	public:
		IpcTimeOutException() {}
		/*!
 					destructor for IpcTimeOutException
		*/
		~IpcTimeOutException () throw() {}
		/*!
 					constructor for IpcTimeOutException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcTimeOutException(const char* reas, const char* fn, int ln)
			: IpcException(reas, fn, ln) {}
		virtual void 		print() const throw();
		virtual void 		print(const char* fn, int ln) const throw();
		/*! 
 			returns class name
		*/
		virtual const char*	classname() const throw() {return "IpcTimeOutException";}
		virtual int			type() {return TIMEOUT;}
	private:
		string format, input;
};

/*!
	IpcConvertException class


	conversion exception class. This type of exception may be thrown by string-to-number methods

*/

class IpcConvertException : public IpcException
{
	public:
		IpcConvertException() {}
		/*!
 					destructor for IpcConvertException
		*/
		~IpcConvertException () throw() {}
		/*!
 					constructor for IpcConvertException

 					in : 
						- reas : exception description   

					in : 
						- fn   : source file name (__FILE__)
					
					in : 
						- ln   : line number in source file (__LINE__)

		*/
		IpcConvertException(const char* reas, const char* fn, int ln, const char* input)
			:  IpcException(reas, fn, ln), remain(input) {}
		virtual void 		print() const throw();
		virtual void 		print(const char* fn, int ln) const throw();
		/*! return string which could not be converted */
		virtual const char*	input() const throw() {return remain.c_str();}
		/*! 
 			returns class name
		*/
		virtual const char*	classname() const throw() {return "IpcConvertException";}
		virtual int			type() {return CONVERT;}
	private:
		string remain;
};

#endif
