
//+=============================================================================
//
// file :         ipc_tcp.cc
//
// description :  C++ source for the classes IPAddr, SocketAdrr, StreamSocket
//				  ListenSocket,ConnectSocket,TcpSocket, ServerSocket 
//                The class is derived from Device. 
//
// project :	  TCP socket encapsulation      
//
// $Author:  	  J. Blume
//
// $Revision:     0.1
//
// $Log:  $
//
// copyleft :    GPL 
//                
// acknowledgments go to   O. Kiselyov (C++ Users Journal, 5/96, 17ff)              
//						   W.R.Stephens, UNP V1
//						   B. Rosenlecher (who taught me a lot about C++)
//=============================================================================



// --- Includes -------------------------------------------------------

#include <sstream>
#include "ipc_tcp.h"
#include "ipc_exception.h"



/*!
 					constructor for SocketAddr, assings parameters
					to base class struct sockaddr_in 

 in: 
	- host :   IPAddr

 in: 
	- port_no:	tcp port number or service name

*/


SocketAddr::SocketAddr(const char* host, const char* port_no)
{

	struct addrinfo *info;
	if((socket_ok = ::getaddrinfo(host, port_no, NULL, &info)) == 0)
	{
		ai_flags 	= info->ai_flags;
		sa_family 	= ai_family = info->ai_family;
		ai_socktype = info->ai_socktype;
		ai_protocol = info->ai_protocol;
		ai_addrlen  = info->ai_addrlen;
		ai_addr		= (sockaddr*) new char[ai_addrlen];
		::memcpy(ai_addr, info->ai_addr, ai_addrlen);
		if(info->ai_canonname)
		{
			if(::strlen(info->ai_canonname) > 0)
			{
				ai_canonname = new char[::strlen(info->ai_canonname)];
				::memcpy(ai_canonname, info->ai_canonname, strlen(info->ai_canonname));
			}
		}
		else
		{
			ai_canonname = NULL;
		}
		::freeaddrinfo(info);
	}
	else
	{
		throw IpcException(::gai_strerror(socket_ok), __FILE__, __LINE__);
	}
}

/*!
 					constructor for SocketAddr, assings parameters
					to base class struct sockaddr_in 

 in: 
	- host :   IPAddr

 in: 
	- port_no:	tcp port number

*/


SocketAddr::SocketAddr(const char* host, unsigned short port_no)
{
	stringstream p;
	p << port_no;
	struct addrinfo *info;
	if((socket_ok = ::getaddrinfo(host, p.str().c_str(), NULL, &info)) == 0)
	{
		ai_flags 	= info->ai_flags;
		sa_family 	= ai_family = info->ai_family;
		ai_socktype = info->ai_socktype;
		ai_protocol = info->ai_protocol;
		ai_addrlen  = info->ai_addrlen;
		ai_addr		= (sockaddr*) new char[ai_addrlen];
		::memcpy(ai_addr, info->ai_addr, ai_addrlen);
		if(info->ai_canonname)
		{
			if(::strlen(info->ai_canonname) > 0)
			{
				ai_canonname = new char[::strlen(info->ai_canonname)];
				::memcpy(ai_canonname, info->ai_canonname, strlen(info->ai_canonname));
			}
		}
		else
		{
			ai_canonname = NULL;
		}
		::freeaddrinfo(info);
	}
	else
	{
		throw IpcException(::gai_strerror(socket_ok), __FILE__, __LINE__);
	}
}

/*!
 					constructor for SocketAddr, assings parameters
					to base class struct sockaddr_in 

 in : 
	- sname :   Unix domain socket device file name


*/

SocketAddr::SocketAddr(const char* sname)
{
	sa_family = ai_family  		= AF_UNIX;
	uds_addr.sun_family 		= ai_family;
	::strcpy(uds_addr.sun_path, sname);
	socket_ok 					= 0;
	
}

/*!

	copy constructor

*/

SocketAddr::SocketAddr(const SocketAddr& in)
{
	socket_ok = in.socket_ok;
	switch(ai_family = in.ai_family)
	{
		case AF_INET:
		case AF_INET6:
			if(socket_ok != 0)
			{
				return;
			}
			ai_flags 	= in.ai_flags;
			sa_family 	= ai_family 	= in.ai_family;
			ai_socktype = in.ai_socktype;
			ai_protocol = in.ai_protocol;
			ai_addrlen  = in.ai_addrlen;
			ai_addr		= (sockaddr*) new char[in.ai_addrlen];
			::memcpy(ai_addr, in.ai_addr, ai_addrlen);
			if(in.ai_canonname)
			{
				if(::strlen(in.ai_canonname) > 0)
				{
					ai_canonname = new char[::strlen(in.ai_canonname)];
					::memcpy(ai_canonname, in.ai_canonname, ::strlen(in.ai_canonname));
				}
			}
		break;
		case AF_UNIX:
			
			uds_addr.sun_family = in.ai_family;
			::strcpy(uds_addr.sun_path, in.uds_addr.sun_path);
		break;
	}
}

SocketAddr::~SocketAddr()
{
	if(ai_family == AF_INET || ai_family == AF_INET6)
	{
		delete[] ai_addr;
	} 
}

/*!
		returns pointer (casted to	sockaddr) to:
		
		
		- struct sockaddr_in (family() == AF_INET)
	
		or

		- sockaddr_in6	(family() == AF_INET6)

		or
	
		- struct sockaddr_un unet (family() == AF_UNIX)


*/

SocketAddr::operator sockaddr* () const
{
	switch(ai_family)
	{
		case AF_INET:
		case AF_INET6:
			return ai_addr;
		case AF_UNIX:
			return (sockaddr*) &uds_addr;
	}
	return NULL;		
}


/*! 

	returns length of appropriate sock_addr structure

*/

socklen_t SocketAddr::size()
{
	switch(ai_family)
	{
		case AF_INET:
		case AF_INET6:
			return ai_addrlen;
			break;
		case AF_UNIX:
			return sizeof(uds_addr);
			break;
	}
	return 0;
}


const char* SocketAddr::info(int flag, int sockfd)
{
	char retname[1024];
	char retserv[1024];

	struct sockaddr addr;
	struct sockaddr* iaddr;
	socklen_t ilen;
	int retval;

	if(sockfd >= 0)
	{
		ilen = sizeof(addr);
		if((retval = ::getsockname(sockfd, (struct sockaddr *)&addr, &ilen)) < 0)
		{
			socket_ok = EAI_AGAIN;
			return 0;
		}
		iaddr = &addr;
	}
	else
	{
		ilen = ai_addrlen;
		iaddr = ai_addr;
	}
	switch(ai_family)
	{
		case AF_INET:
		case AF_INET6:
		switch(flag)
		{
			case IPC_TCPNAME:
				if(!(socket_ok = ::getnameinfo(iaddr, ilen, retname, 1024, retserv, 1024, 0)))
				{	
					sockInfo = retname;
				}
				break;
			case IPC_TCPADDR:
				if(!(socket_ok = ::getnameinfo(iaddr, ilen, retname, 1024, retserv, 1024, NI_NUMERICHOST)))
				{
					sockInfo = retname;
				}
				break;
			case IPC_TCPPORTNAME:
				if(!(socket_ok = ::getnameinfo(iaddr, ilen, retname, 1024, retserv, 1024, 0)))
				{
					sockInfo = retserv;
				}
				break;
			case IPC_TCPPORTNUM:
				if(!(socket_ok = ::getnameinfo(iaddr, ilen, retname, 1024, retserv, 1024, NI_NUMERICSERV)))
				{
					sockInfo = retserv;
				}
				break;
		}
		break;
		case AF_UNIX:
			sockInfo = uds_addr.sun_path;
		break;
	}
	return socket_ok ? 0 : sockInfo.c_str();
}

/*!						Constructor for ConnectSocket, connects to targetAddr

	in: 
		- targetAddr:	Adress and port of the server to connect to
	in: 
		- timeout:      Time in secs to wait for successful connect

	Can throw:

		- IpcException
*/



ConnectSocket::ConnectSocket(SocketAddr targetAddr, long timeout) :  storedAddress(targetAddr)
{
	int				flags, n;
	struct timeval	tval;
	
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	if( (fd = ::socket(storedAddress.family(), SOCK_STREAM, 0)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if( (flags = ::fcntl(fd, F_GETFL, 0)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}

	if( (::fcntl(fd, F_SETFL, flags | O_NONBLOCK)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}

	if(storedAddress.family() == AF_INET || storedAddress.family() == AF_INET6)
	{
		if(!storedAddress.ok())
		{
			throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
		}
	}
	if ( (n = ::connect(fd, (sockaddr*) storedAddress, storedAddress.size())) < 0)
	{
		if (errno != EINPROGRESS)
		{
		 	throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		if(n == 0 && (::fcntl(fd, F_SETFL, flags)) != flags)
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
	}
	// check for connection success	
	if(storedAddress.family() == AF_INET || storedAddress.family() == AF_INET6)
	{
		int 	optval;
		fd_set	wdfs;
		socklen_t optvallen = sizeof(optval);
		FD_ZERO(&wdfs);
		FD_SET(fd, &wdfs);
		tval.tv_sec = timeout;
		tval.tv_usec = 0;
		if( (n = ::select(fd + 1, NULL, &wdfs, NULL, &tval)) < 1)
		{
			if(n == 0)
			{
				errno = ETIMEDOUT;
			}
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		if( (n = ::getsockopt(fd, SOL_SOCKET, SO_ERROR, &optval, &optvallen)) != 0)
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		if(optval != 0)
		{
			throw IpcException("Connection failed", __FILE__, __LINE__);
		}
	}
	if(::fcntl(fd, F_SETFL, flags) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
}


/*!			returns: 
 * 			- the address of the host the object is connected to
 *
 			Can throw:

				- IpcException

			

*/
const char* ConnectSocket::hostAddress()
{
	const char* address = storedAddress.info(IPC_TCPADDR, fd);
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return address;
}


/*!		returns:
 * 			- the name of the host the object is connected to
 *
		Can throw:

				- IpcException

*/

const char* ConnectSocket::hostName()
{
	const char* name = storedAddress.info(IPC_TCPNAME, fd);
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return name;
}

/*!					returns: 
 * 					- the name of the port the object is connecting from
 *
 *
		Can throw:

				- IpcException
*/

const char* ConnectSocket::hostPname()
{
	const char* name = storedAddress.info(IPC_TCPPORTNAME, fd);
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return name;

}


/*!			
			returns:  
					- the ip address of the host the object is connecting to

			Can throw:

			- IpcException
*/
const char* ConnectSocket::peerAddress()
{
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return storedAddress.address();
}

/*!			
			returns: 
			- the name of the host the object is connecting to

			Can throw:

			- IpcException
*/
const char* ConnectSocket::peerName()
{
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return storedAddress.name();
}

/*!			
		returns: 
			- the name of the port the object is connecting to

		Can throw:

			- IpcException

*/
		
const char* ConnectSocket::peerPname()
{
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return storedAddress.pname();
}

/*!			
		returns: 
				- the number of the port the object is connecting to

		Can throw:

			- IpcException
*/
		
const char* ConnectSocket::peerPnum()
{
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return storedAddress.pnum();
}


/*!
			shuts down all or part of the network connection, cf. man 2 shutdown

			in:	
				- SHUT_RD, SHUT_WR or SHUT_RDWR, cf. man 2 shutdown\n
					
			returns: 
				- the return value of std::shutdown(int how)

			Can throw:

				- IpcException	

*/

int ConnectSocket::shutdown(int how)
{
	int retval;
	if((retval = ::shutdown(fd, how)) != 0) 
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
    return retval;
}

/*!
  closes the connection and opens it again

  returns: 
	- the socket descriptor of the reopened connection

  (if using the class with Io remember to tell Io about the renewed connetion using Io::client())

	Can throw:

				- IpcException
*/

int ConnectSocket::reopen()
{
	int retval = -1, flags;
	::close(fd);
	if((fd = ::socket(storedAddress.family(), SOCK_STREAM, 0) < 0))
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if( (flags = ::fcntl(fd, F_GETFL, 0)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if ( (retval = ::connect(fd, (sockaddr*) storedAddress, storedAddress.size())) < 0)
	{
		if (errno != EINPROGRESS)
		{
			 throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
		if(retval == 0 && (::fcntl(fd, F_SETFL, flags)) != flags)
		{
			throw IpcException(::strerror(errno), __FILE__, __LINE__);
		}
	}
	return retval;
}

/*!					

	returns: 
		-the number of the port the object is connected or
	
	Can throw:

				- IpcException
*/


const char* ConnectSocket::hostPnum()
{
	const char* num = storedAddress.info(IPC_TCPPORTNUM, fd);
	if(!storedAddress.ok())
	{
		throw IpcException(storedAddress.socketErr(), __FILE__, __LINE__);
	}
	return num;
}
/*!
					sets/unsets TCP_NODELAY option for the socket (may speed up throughput considerably)

					in: 
						- true: the TCP_NODELAY will be turned on
					      
                        -  false:  the TCP_NODELAY will be turned off
					
					returns: 
						
						- true on success, false on erro

					Can throw:

						- IpcException

					this only makes sense for TcpSocket class, for TtyDevice
					return value is always true
*/
bool ConnectSocket::nodelay(bool on)
{
	int retval, flag;

	on ? flag = 1 : flag = 0;
	if((retval = ::setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) )) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	return true;
}
/*!
					reports TCP_NODELAY option state for the socket

					return: 
							- true: 	TCP_NODELAY is enabled

							- false:	TCP_NODELAY is disabled or an error occured, so check with ok()

					Can throw:

						- IpcException


					
					this only makes sense for TcpSocket class, for TtyDevice
					return value is always true
*/

bool ConnectSocket::nodelay()
{
	int retval, flag;
	socklen_t len;

	len = sizeof(int);
	if((retval = ::getsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, &len)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if(len != sizeof(int))
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if(flag == 1)
	{
		return true;
	}
	return false;
}

/*!
					reports  SO_KEEPALIVE option state for the socket

					return: 
							- true: 	SO_KEEPALIVE is enabled

							- false:	SO_KEEPALIVE is disabled

					Can throw:

						- IpcException


					
*/

bool ConnectSocket::keepalive()
{
	int retval, flag;
	socklen_t len;

	len = sizeof(int);
	if((retval = ::getsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *) &flag, &len)) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if(len != sizeof(int))
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
	if(flag == 1)
	{
		return true;
	}
	return false;
}

/*!
					sets/unsets SO_KEEPALIVE option for the socket (needed sometimes to prevent a peer from closing
                    the connection when there's no activity for longer periods of time

					in: 
						- true: the SO_KEEPALIVE will be turned on
					      
                        -  false:  the SO_KEEPALIVE  will be turned off
					
					returns: 
						
						- true on success, false on erro

					Can throw:

						- IpcException

*/

bool ConnectSocket::keepalive(bool on, int keepalive_interval)
{
	int retval, flag;

	on ? flag = 1 : flag = 0;
	if((retval = ::setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&flag, sizeof(flag) )) < 0)
	{
		throw IpcException(::strerror(errno), __FILE__, __LINE__);
	}
    if(on)
    {
        if((retval = ::setsockopt(fd, SOL_TCP, TCP_KEEPINTVL, (char *)&keepalive_interval, sizeof(keepalive_interval) )) < 0)
	    {
		    throw IpcException(::strerror(errno), __FILE__, __LINE__);
	    }
    }
	return true;
}
/*!
	


	  Constructor for TcpSocket, creates a connection using tcp socket
//
// in: 
// 	-  host:		hostname / IP address to connect to

// in: 
// 	-  service: 	portnumber or service name to connect to

// in: 
// 	-  timeout	timeout in sec to wait for connect (defaults to 2)

// Can throw:
// 	- IpcException					
//					
//
*/

TcpSocket::TcpSocket(const char* host, const char* service, long timeout) 
	: ConnectSocket(SocketAddr(host, service), timeout)
{

}

/*!					Constructor for TcpSocket, creates a connection using tcp socket
//
// in: 
// 	-  host:		hostname / IP address to connect to

// in: 
// 	-  port_no:		portnumber to connect to

// in: 
// 	-  timeout	timeout in sec to wait for connect (defaults to 2)

// Can throw:
// 	- IpcException					
//					
//
*/
TcpSocket::TcpSocket(const char* host, unsigned short port_no, long timeout)
	: ConnectSocket(SocketAddr(host, port_no), timeout)
{
}

/*!					Constructor for DomainSocket, creates a connection using unix domain socket
//
// in: 
// -  sname		socket filename


// in: 
// -  timeout	timeout in sec to wait for connect (defaults to 2)
// 
//  Can throw:
// 	- IpcException					
*/

DomainSocket::DomainSocket(const char* sname, long timeout) 
	: ConnectSocket(SocketAddr(sname), timeout)
{

}


//--- ende ipc_tcp.cc ------------------------------------------------- 

