#include <iostream>
#include "ipc_tty.h"
#include "ipc_exception.h"

/*!
	constructor for class TtyDevice

	After successfull construction the object represents an
	opened tty device.


	- tty 	:  the device file name
	- speed :  the baud rate
	- csize :  the character size mask (5, 6, 7 or 8)
	- parity:  0 = off, 1 = odd, 2 = even
	- stopbit: stop bit (0, 1 or 2)
	- flowcontrol:  0 = none, 1 = software, 2 = hardware

	Can throw:
		- IpcException
		
*/

TtyDevice::TtyDevice(const char *tty, long speed, int csize, int parity, int stopbit, int flowctrl)
	: device(tty)
{
	init_speeds();
	if((fd = tty_open(device.c_str())) > -1)
	{
		if(tty_get_settings(&old_term_ios) > -1)
		{
			if(tty_set_speed(long_to_speed(speed)) > -1)
			{
				if(tty_set_csize(csize) > -1)
				{
					if(tty_set_parity(parity) > -1)
					{
						if(tty_set_stop_bits(stopbit) > -1)
						{
							if(set_xon_in(flowctrl) > -1)
							{
								if(set_xon_out(flowctrl) > -1)
								{
									tty_get_settings(&term_ios);
								}
							}
						}
					}
				}
			}
		}
	}
}

TtyDevice::~TtyDevice()
{
	vector<speed_setting*>::iterator it;
	for(it = speeds.end() - 1; it >= speeds.begin(); it--)
	{
		delete *it;
	}	
}

void TtyDevice::init_speeds()
{
	speeds.push_back(new speed_setting(B0, 0));
	speeds.push_back(new speed_setting(B50, 50));
	speeds.push_back(new speed_setting(B75, 75));
	speeds.push_back(new speed_setting(B110, 110));
	speeds.push_back(new speed_setting(B134, 134));
	speeds.push_back(new speed_setting(B150, 150));
	speeds.push_back(new speed_setting(B200, 200));
	speeds.push_back(new speed_setting(B300, 300));
	speeds.push_back(new speed_setting(B600, 600));
	speeds.push_back(new speed_setting(B1200, 1200));
	speeds.push_back(new speed_setting(B1800, 1800));
	speeds.push_back(new speed_setting(B2400, 2400));
	speeds.push_back(new speed_setting(B4800, 4800));
	speeds.push_back(new speed_setting(B9600, 9600));
	speeds.push_back(new speed_setting(B19200, 19200));
	speeds.push_back(new speed_setting(B38400, 38400));
	speeds.push_back(new speed_setting(B57600, 57600));
	speeds.push_back(new speed_setting(B115200, 115200));
	speeds.push_back(new speed_setting(B230400, 230400));
}

speed_t TtyDevice::long_to_speed(long speed)
{
	vector<speed_setting*>::iterator it;
	for(it = speeds.begin(); it < speeds.end(); it++)
	{
		if((*it)->get_l_speed() == speed)
		{
			return (*it)->get_s_speed();
		}
	}
	stringstream s;
	s << "Invalid speed setting: " << speed;
	throw IpcException(s.str(),  __FILE__, __LINE__);
}

long TtyDevice::speed_to_long(speed_t speed)
{
	vector<speed_setting*>::iterator it;
	for(it = speeds.begin(); it < speeds.end(); it++)
	{
		if((*it)->get_s_speed() == speed)
		{
			return (*it)->get_l_speed();
		}
	}
	stringstream s;
	s << "Invalid speed setting: " << speed;
	throw IpcException(s.str(),  __FILE__, __LINE__);
}

int TtyDevice::tty_open(const char* tty)
{
	if((fd = ::open(tty, O_RDWR)) < 0)
    {
		throw IpcException(::strerror(errno),  __FILE__, __LINE__);
    }
	return fd;
}

int TtyDevice::tty_get_settings(struct termios* tios)
{
	if(tcgetattr(fd, tios) < 0)
    {
		throw IpcException(::strerror(errno),  __FILE__, __LINE__);
    }
	return fd;
}

int TtyDevice::tty_set_settings(struct termios* tios)
{
	if(tcsetattr(fd, TCSANOW, tios) < 0)
    {
		throw IpcException(::strerror(errno),  __FILE__, __LINE__);
    }
	return fd;
}

int TtyDevice::tty_set_speed(speed_t speed)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		::cfmakeraw(&term_ios);
    	::cfsetispeed(&term_ios, speed);
    	::cfsetospeed(&term_ios, speed);
		if(tty_set_settings(&term_ios) < 0)
    	{
			throw IpcException(::strerror(errno),  __FILE__, __LINE__);
    	}
		return 0;
	}
	return -1; // sedate compiler	
}

long TtyDevice::tty_get_speed()
{
	struct termios term_ios;
	
	if(tty_get_settings(&term_ios) > -1)
	{
		if(cfgetispeed(&term_ios) == cfgetospeed(&term_ios))
		{
			return speed_to_long(cfgetispeed(&term_ios));
		}
		else
		{
			throw IpcException("Different settings for input and output speed discovered", __FILE__, __LINE__);
		}
	}
	return -1;
}


int TtyDevice::tty_set_csize(int cs)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		switch(cs)
		{
			case 5:
				term_ios.c_cflag &= ~(CSIZE);
				term_ios.c_cflag |= CS5;
				break;
			case 6:
				term_ios.c_cflag &= ~(CSIZE);
				term_ios.c_cflag |= CS6;
				break;
			case 7:
				term_ios.c_cflag &= ~(CSIZE);
				term_ios.c_cflag |= CS7;
				break;
			case 8:
				term_ios.c_cflag &= ~(CSIZE);
				term_ios.c_cflag |= CS8;
				break;
			default:
				stringstream s;
				s << "Invalid character size mask setting: " << cs;
				throw IpcException(s.str(), __FILE__, __LINE__);
		}
		return tty_set_settings(&term_ios);
	}
	return -1;
}

int TtyDevice::tty_get_csize()
{
	int c;
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		switch(term_ios.c_cflag & CSIZE)
		{
			case CS5:
				c = 5;
				break;
			case CS6:
				c = 6;
				break;
			case CS7:
				c = 7;
				break;
			case CS8:
				c = 8;
				break;
			default:
				stringstream s;
				s << "Invalid character size mask setting: " << (term_ios.c_cflag & CSIZE);
				throw IpcException(s.str(), __FILE__, __LINE__);
				break;
		}
		return c;
	}
	return -1;
}

int TtyDevice::tty_set_parity(int on)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(on == 1)
		{
			term_ios.c_cflag |= PARENB;
			term_ios.c_cflag |= PARODD;
			
		}
		else if(on == 2)
		{
			term_ios.c_cflag |= PARENB;
			term_ios.c_cflag &= ~ PARODD;
			
		}
		else
		{
			term_ios.c_cflag &= ~ PARENB;			
		}
		if( tty_set_settings(&term_ios) < 0)
		{
			throw IpcException("Error writing parity setting", __FILE__, __LINE__);
		}
		return fd;
	}
	return -1;	
}

bool TtyDevice::tty_get_parity()
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		return bool (term_ios.c_cflag & PARENB);
	}
	return false; // sedate compiler
}

int TtyDevice::tty_set_stop_bits(int num)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(num == 2)
		{
			term_ios.c_cflag |= CSTOPB;
		}
		else if(num == 1)
		{
			term_ios.c_cflag &= ~ CSTOPB;
		}
		else
		{
			stringstream s;
			s << "Invalid number of stop bits: " << num;
			throw IpcException(s.str(), __FILE__, __LINE__);
		}
		if(tty_set_settings(&term_ios) < 0)
		{
			throw IpcException("Error writing stop bit setting", __FILE__, __LINE__);
		}
		return fd;
	}
	return -1;
}

int TtyDevice::tty_get_stop_bits()
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(term_ios.c_cflag & CSTOPB)
		{
			return 2;
		}
		else
		{
			return 1;
		}
	}
	throw IpcException("Error reading stop bits", __FILE__, __LINE__);
}

int TtyDevice::tty_set_xon_in(int on)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(on > 0)
		{
			term_ios.c_iflag |= IXON;
			if(on > 1)
			{
				term_ios.c_cflag |= CRTSCTS;
			}
			else
			{
				term_ios.c_cflag &= ~CRTSCTS;
			}
		}
		else
		{
			term_ios.c_iflag &= ~IXON;
			term_ios.c_cflag &= ~CRTSCTS;
		}
		if( tty_set_settings(&term_ios) < 0)
		{
			throw IpcException("Error writing flow control settings", __FILE__, __LINE__);
		}
		return fd;
	}
	else
	{
		throw IpcException("Error reading tty settings", __FILE__, __LINE__);
	}
	return -1;
}

int	TtyDevice::tty_set_xon_out(int on)
{
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(on > 0)
		{
			term_ios.c_iflag |= IXOFF;
			if(on > 1)
			{
				term_ios.c_cflag |= CRTSCTS;
			}
			else
			{
				term_ios.c_cflag &= ~CRTSCTS;
			}
		}
		else
		{
			term_ios.c_iflag &= ~IXOFF;
			term_ios.c_cflag &= ~CRTSCTS;
		}
		if( tty_set_settings(&term_ios) < 0)
		{
			throw IpcException("Error writing flow control settings", __FILE__, __LINE__);
			return -1;
		}
		return fd;
	}
	return -1;
}

void TtyDevice::set_flowcontrol(int on)
{
	tty_set_xon_in(on);
	tty_set_xon_out(on);
}

int TtyDevice::get_xon_in()
{
	int retval = 0;
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(term_ios.c_iflag & IXON)
		{
			++retval;
			if(term_ios.c_cflag & CRTSCTS)
			{
				++retval;
			}
		}
	}
	else
	{
		throw IpcException("Error reading tty settings", __FILE__, __LINE__);
	}
	return retval;
}

int TtyDevice::get_xon_out()
{
	int retval = 0;
	struct termios term_ios;
	if(tty_get_settings(&term_ios) > -1)
	{
		if(term_ios.c_oflag & IXON)
		{
			++retval;
			if(term_ios.c_cflag & CRTSCTS)
			{
				++retval;
			}
		}
	}
	else
	{
		throw IpcException("Error reading tty settings", __FILE__, __LINE__);
	}
	return retval;
}

int TtyDevice::get_flowcontrol()
{
	int val1 = get_xon_in();
	int val2 = get_xon_out();
	if(val1 != val2)
	{
		throw IpcException("Flow control input / output settings differ", __FILE__, __LINE__);
	}
	return val1;
}
