#ifndef IPC_TTY_H
#define IPC_TTY_H

#include <vector>
#include <string>
#include <sstream>
#include <cstring>

using namespace std;

extern "C"
{
	#include <unistd.h>
	#include <sys/types.h>
    #include <sys/stat.h>
    #include <fcntl.h>
	#include <errno.h>
	#include <termios.h>
}


/*!
	class speed_setting\n

	class used to map integer baud rate values to termios constants
 
*/

class speed_setting
{
	public:
		/*!
			constructor for speed_setting
			
			- s_speed : the termios baud rate constant
			- l_speed : the corresponding baud rate integer value
		*/
		speed_setting(speed_t s,long l) {s_speed =s; l_speed = l;}
		/*!
			returns the termios baud rate constant
		*/
		speed_t get_s_speed() {return s_speed;}
		/*!
			returns the baud rate integer value
		*/
		long get_l_speed() {return l_speed;}
	private:
		speed_t s_speed;
		long	l_speed;
};


/*!
	class TTyDevice\n

	class represents a serial device
 
*/

class TtyDevice 
{
	public:
		TtyDevice(const char* tty, long speed, int csize = 8 , int parity = 0 , int stopbit = 1, int flowctrl = 0);
		~TtyDevice();

		/*!
 * 			returns tty device name
 * 		*/
		const char* deviceName() {return device.c_str();}
		/*!
 			returns file descriptor of established connection
		*/
		int	filedescriptor() {return fd;}
		/*!
					returns error status of last operation on TtyDevice

					true:  error\n
					false: no error

					calling this method will reset the error state, i.e. a second call will
					return true if no further error occured between the two calls
		*/

		//bool ok() {return glErr.ok();}
		/*!
 
				    returns the latest error message

					calling this method will clear the error message buffer,\n
					i.e. a second call will return an empty string if no further\n
					error occured between the two calls	
		*/

		//const char* getError()				{return glErr.last_err();}

		/*!
			returns the current character size mask
		*/
		int			get_csize() 				{return tty_get_csize();}
		/*!
			set the character size mask to c (5, 6, 7 or 8)
		*/
		int			set_csize(int c) 			{return tty_set_csize(c);}
		/*!
			returns the current baud rate setting as a long value
		*/
		long 		get_speed()					{return tty_get_speed();}
		/*!
			sets the  baud rate to speed (termios constant, cd. man termios)
		*/
		int			set_speed(speed_t speed)	{return tty_set_speed(speed);}
		/*!
			returns the current parity setting, true = enabled, false = disabled
		*/
		bool		get_parity() 				{return tty_get_parity();}
		/*!
			sets parity to on, true = enabled, false = disabled
		*/
		int			set_parity(bool on) 		{return tty_set_parity(on);}
		/*!
			returns the current stop bit value
		*/
		int			get_stop_bits() 			{return  tty_get_stop_bits();}
		/*!
			sets stop bit value to num
		*/
		int			set_stop_bits(int num) 		{return tty_set_stop_bits(num);}
		
		/*! 
			set flow control  on input, 0 = none, 1 = software, 2 = hardware
		*/
		int			set_xon_in(int on)			{return tty_set_xon_in(on);}

		/*! 
			set flow control on output, 0 = none, 1 = software, 2 = hardware
		*/
		int			set_xon_out(int on)			{return tty_set_xon_out(on);}
		/*! 
			set flow control on input and output, 0 = none, 1 = software, 2 = hardware
		*/
		void	set_flowcontrol(int on);
		/*! 
			get flow control  on input, 0 = none, 1 = software, 2 = hardware
		*/		
		int get_xon_in();
		/*! 
			get flow control  on output, 0 = none, 1 = software, 2 = hardware
		*/
		int get_xon_out();
		/*! 
			get flow control , 0 = none, 1 = software, 2 = hardware
		*/
		int get_flowcontrol();
		


	private:
		void init_speeds();
		speed_t long_to_speed(long speed);
		long speed_to_long(speed_t speed);
		int tty_open(const char* tty);
		int tty_get_settings(struct termios* tios);
		int tty_set_settings(struct termios* tios);
		int tty_set_speed(speed_t speed);
		long tty_get_speed();
		int tty_set_csize(int cs);
		int tty_get_csize();
		int tty_set_parity(int on);
		bool tty_get_parity();
		int tty_set_stop_bits(int num);
		int tty_get_stop_bits();
		int tty_set_xon_in(int on);
		int tty_set_xon_out(int on);
		bool flush(int queue);
		struct termios old_term_ios, term_ios;
		int fd;
		string 	device;
		vector<speed_setting*> speeds;
};

#endif
